@@ -571,6 +571,30 @@ f_sqrt(acb_ptr res, const acb_t z, void * param, slong order, slong prec)
     return 0;
 }
 
+/* f(z) = 1/sqrt(z) */
+int
+f_rsqrt(acb_ptr res, const acb_t z, void * param, slong order, slong prec)
+{
+    if (order > 1)
+        flint_abort();  /* Would be needed for Taylor method. */
+
+    acb_rsqrt_analytic(res, z, order != 0, prec);
+
+    return 0;
+}
+
+/* f(z) = rgamma(z) */
+int
+f_rgamma(acb_ptr res, const acb_t z, void * param, slong order, slong prec)
+{
+    if (order > 1)
+        flint_abort();  /* Would be needed for Taylor method. */
+
+    acb_rgamma(res, z, prec);
+
+    return 0;
+}
+
 /* f(z) = exp(-z^2+iz) */
 int
 f_gaussian_twist(acb_ptr res, const acb_t z, void * param, slong order, slong prec)
@@ -628,11 +652,88 @@ f_sin_cos_frac(acb_ptr res, const acb_t z, void * param, slong order, slong prec
     return 0;
 }
 
+/* f(z) = sin((1/1000 + (1-z)^2)^(-3/2)), example from Mioara Joldes' thesis
+                                          (suggested by Nicolas Brisebarre) */
+int
+f_sin_near_essing(acb_ptr res, const acb_t z, void * param, slong order, slong prec)
+{
+    acb_t t, u;
+
+    if (order > 1)
+        flint_abort();  /* Would be needed for Taylor method. */
+
+    acb_init(t);
+    acb_init(u);
+
+    acb_sub_ui(t, z, 1, prec);
+    acb_neg(t, t);
+    acb_mul(t, t, t, prec);
+    acb_one(u);
+    acb_div_ui(u, u, 1000, prec);
+    acb_add(t, t, u, prec);
+    acb_set_d(u, -1.5);
+    acb_pow_analytic(t, t, u, order != 0, prec);
+    acb_sin(res, t, prec);
+
+    acb_clear(t);
+    acb_clear(u);
+
+    return 0;
+}
+
+/* f(z) = exp(-z) (I_0(z/k))^k, from Bruno Salvy */
+int
+f_scaled_bessel(acb_ptr res, const acb_t z, void * param, slong order, slong prec)
+{
+    acb_t nu;
+
+    ulong k = ((ulong *) param)[0];
+
+    acb_init(nu);
+    acb_div_ui(res, z, k, prec);
+    acb_hypgeom_bessel_i_scaled(res, nu, res, prec);
+    acb_pow_ui(res, res, k, prec);
+    acb_clear(nu);
+
+    return 0;
+}
+
+/*
+Bound for scaled Bessel function: 2/(2 pi x)^(1/2)
+Bound for tail of integral: 2 N (k / (pi N))^(k / 2) / (k - 2).
+*/
+void
+scaled_bessel_tail_bound(arb_t b, ulong k, const arb_t N, slong prec)
+{
+    arb_const_pi(b, prec);
+    arb_mul(b, b, N, prec);
+    arb_ui_div(b, k, b, prec);
+    arb_sqrt(b, b, prec);
+    arb_pow_ui(b, b, k, prec);
+    arb_mul(b, b, N, prec);
+    arb_mul_ui(b, b, 2, prec);
+    arb_div_ui(b, b, k - 2, prec);
+}
+
+void
+scaled_bessel_select_N(arb_t N, ulong k, slong prec)
+{
+    slong e;
+    double f = log(k/3.14159265358979)/log(2);
+
+    e = 1;
+    while ((k / 2.0) * (e - f) - e < prec + 5)
+        e++;
+
+    arb_one(N);
+    arb_mul_2exp_si(N, N, e);
+}
+
 /* ------------------------------------------------------------------------- */
 /*  Main test program                                                        */
 /* ------------------------------------------------------------------------- */
 
-#define NUM_INTEGRALS 32
+#define NUM_INTEGRALS 37
 
 const char * descr[NUM_INTEGRALS] =
 {
@@ -668,14 +769,21 @@ const char * descr[NUM_INTEGRALS] =
     "int_0^{inf} exp(-x^2+ix) dx   (using domain truncation)",
     "int_0^{inf} exp(-x) Ai(-x) dx   (using domain truncation)",
     "int_0^pi x sin(x) / (1 + cos(x)^2) dx",
+    "int_0^3 sin(0.001 + (1-x)^2)^(-3/2)) dx  (slow convergence, use higher -eval)",
+    "int_0^{inf} exp(-x) I_0(x/3)^3 dx   (using domain truncation)",
+    "int_0^{inf} exp(-x) I_0(x/15)^{15} dx   (using domain truncation)",
+    "int_{-1-i}^{-1+i} 1/sqrt(x) dx",
+    "int_0^{inf} 1/gamma(x) dx   (using domain truncation)",
 };
 
 int main(int argc, char *argv[])
 {
     acb_t s, t, a, b;
     mag_t tol;
+    slong num_threads;
     slong prec, goal;
     slong N;
+    ulong k;
     int integral, ifrom, ito;
     int i, twice, havegoal, havetol;
     acb_calc_integrate_opt_t options;
@@ -714,7 +822,8 @@ int main(int argc, char *argv[])
         flint_printf("-verbose2  - show more information\n");
         flint_printf("-deg n     - use quadrature degree up to n\n");
         flint_printf("-eval n    - limit number of function evaluations to n\n");
-        flint_printf("-depth n   - limit subinterval queue size to n\n\n");
+        flint_printf("-depth n   - limit subinterval queue size to n\n");
+        flint_printf("-threads n - use parallel computation with n threads\n\n");
         flint_printf("Implemented integrals:\n");
         for (integral = 0; integral < NUM_INTEGRALS; integral++)
             flint_printf("I%d = %s\n", integral, descr[integral]);
@@ -728,6 +837,7 @@ int main(int argc, char *argv[])
     twice = 0;
     goal = 0;
     havetol = havegoal = 0;
+    num_threads = 1;
 
     acb_init(a);
     acb_init(b);
@@ -788,8 +898,15 @@ int main(int argc, char *argv[])
         {
             options->use_heap = 1;
         }
+        else if (!strcmp(argv[i], "-threads"))
+        {
+            num_threads = atol(argv[i+1]);
+        }
     }
 
+    if (num_threads >= 2)
+        flint_set_num_threads(num_threads);
+
     if (!havegoal)
         goal = prec;
 
@@ -1112,6 +1229,46 @@ int main(int argc, char *argv[])
                 acb_calc_integrate(s, f_sin_cos_frac, NULL, a, b, goal, tol, options, prec);
                 break;
 
+            case 32:
+                acb_zero(a);
+                acb_set_ui(b, 3);
+                acb_calc_integrate(s, f_sin_near_essing, NULL, a, b, goal, tol, options, prec);
+                break;
+
+            case 33:
+                acb_zero(a);
+                acb_zero(b);
+                k = 3;
+                scaled_bessel_select_N(acb_realref(b), k, prec);
+                acb_calc_integrate(s, f_scaled_bessel, &k, a, b, goal, tol, options, prec);
+                scaled_bessel_tail_bound(acb_realref(a), k, acb_realref(b), prec);
+                arb_add_error(acb_realref(s), acb_realref(a));
+                break;
+
+            case 34:
+                acb_zero(a);
+                acb_zero(b);
+                k = 15;
+                scaled_bessel_select_N(acb_realref(b), k, prec);
+                acb_calc_integrate(s, f_scaled_bessel, &k, a, b, goal, tol, options, prec);
+                scaled_bessel_tail_bound(acb_realref(a), k, acb_realref(b), prec);
+                arb_add_error(acb_realref(s), acb_realref(a));
+                break;
+
+            case 35:
+                acb_set_d_d(a, -1, -1);
+                acb_set_d_d(b, -1, 1);
+                acb_calc_integrate(s, f_rsqrt, NULL, a, b, goal, tol, options, prec);
+
+            case 36:
+                if (goal < 0)
+                    abort();
+                acb_zero(a);
+                acb_set_ui(b, 4 + (goal + 1) / 2);
+                acb_calc_integrate(s, f_rgamma, NULL, a, b, goal, tol, options, prec);
+                arb_add_error_2exp_si(acb_realref(s), -goal);
+                break;
+
             default:
                 abort();
             }
@@ -1129,7 +1286,7 @@ int main(int argc, char *argv[])
     acb_clear(t);
     mag_clear(tol);
 
-    flint_cleanup();
+    flint_cleanup_master();
     return 0;
 }
 
